#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

from __future__ import with_statement
from __future__ import unicode_literals

import os
import sys
import stat
import time
import calendar
import re
import json
import ssl

try:
    import box.urllib2 as urllib2
except:
    import urllib2

try:
    import box.httplib as httplib
except:
    import httplib

from urllib import (urlencode,)
from urlparse import (urlparse, urlunparse,)

from fileaccess import (FileObj, Access)

from pathid import (PathIdMap,)
from util import (to_utf8, get_stream, DataStream, UploadFileData)

from pkgaddon.box_pkg_config import (ROOT_PATH, ROOT_FOLDER_ID,)


OPEN_MODE_READ = "r"
OPEN_MODE_WRITE = "w"

SESSION_PATH_PERM = (stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP |
                        stat.S_IWGRP | stat.S_IROTH | stat.S_IWOTH)

ACCESS_TOKEN_NAME = "access_token"
REFRESH_TOKEN_NAME = "refresh_token"
TOTAL_COUNT_NAME = "total_count"
ENTRIES_NAME = "entries"
TYPE_NAME = "type"
ID_NAME = "id"
NAME_NAME = "name"
SIZE_NAME = "size"
MODIFIED_AT_NAME = "modified_at"
PARENT_NAME = "parent"
CONTEXT_INFO_NAME = "context_info"
CONFLICTS_NAME = "conflicts"
CODE_NAME = "code"

FOLDER_TYPE = "folder"
FILE_TYPE = "file"

FOLDERS_ITEMS_PARAM_TMPL = \
    "fields=type,id,name,size,modified_at&limit=%d&offset=%d"
RECURSIVE_PARAM = "recursive=true"
ITEM_LIMIT = 1000

FOLDERS_URL_TMPL = "/folders"
FOLDERS_ID_URL_TMPL = "/folders/%s"
FOLDERS_ITEMS_URL_TMPL = FOLDERS_ID_URL_TMPL + "/items"
MAKE_DIRECTORY_URL_TMPL = FOLDERS_URL_TMPL
FILES_URL_TMPL = "/files"
FILES_ID_URL_TMPL = "/files/%s"
FILE_READ_URL_TMPL = FILES_ID_URL_TMPL + "/content"
FILE_WRITE_URL_TMPL = FILES_URL_TMPL + "/content"
FILE_UPDATE_URL_TMPL = FILES_ID_URL_TMPL + "/content"

POST_METHOD = "POST"
PUT_METHOD = "PUT"
DELETE_METHOD = "DELETE"

BOUNDARY_MAIN = "BOX_FILE_UPLOAD_BOUDARY"

UPLOAD_ATTRIBUTES_TMPL = '{"name": "%s", "parent": {"id": "%s"}}'

MTIME_UTC_RE = "(.*)Z"
MTIME_TZ_RE = "(.*)([+-])(\d\d):*(\d\d)"
MTIME_FORMAT = "%Y-%m-%dT%H:%M:%S"

UNAUTHORIZED_RETRY_COUNT = 1
SERVER_ERR_RETRY_COUNT = 3
RETRY_SLEEP_TIME = 30

ERR_MSG_INVALID_OPEN_MODE = "error invalid open mode"

SSL_CREATE_UNVERIFIED_CONTEXT = "_create_unverified_context"

PROTOCOL_LIST = ("http", "https")


def to_json(response):
    """Load JSON data from "response" and return it."""

    json_response = {}
    try:
        json_response = json.loads(response)
    except:
        pass
    return json_response


def recreate_http_error(http_err, err_data):
    """Recreate urllib2.HTTPError instance and return it."""

    return urllib2.HTTPError(http_err.filename, http_err.code, http_err.msg,
                              http_err.hdrs, get_stream(err_data))


def get_proxies_dict(proxy, protocol_list):
    """Return proxies for urllib2.ProxyHandler."""

    if not proxy:
        return None

    parse_result = urlparse(proxy)
    if parse_result.scheme and parse_result.netloc:
        proxy = parse_result.netloc
    proxies = {}
    for protocol in protocol_list:
        proxies[protocol] = \
                    urlunparse((protocol, proxy, "", "", "", ""))

    if not proxies:
        proxies = None
    return proxies


class NeedToRetry:
    """Class for representing that retry is needed."""

    pass


class ErrorInfo(object):
    """Class for defining the error code."""

    NOT_FOUND = "not_found"
    ITEM_NAME_IN_USE = "item_name_in_use"

    def __init__(self, json_response):

        try:
            self._code = json_response[CODE_NAME]
        except:
            self._code = ""

        try:
            self._id = \
                json_response[CONTEXT_INFO_NAME][CONFLICTS_NAME][ID_NAME]
        except:
            self._id = ""

    @property
    def code(self):

        return self._code

    @property
    def id(self):

        return self._id


class BoxUploadFileData(UploadFileData):
    """Class for upload file data for BOX."""

    def _get_form_data_string(self, name, data=None, type=None, file_name=""):
        """Return the prefix string data for uploading."""

        form_data = to_utf8(self.BOUNDARY_PREFIX + self.boundary + \
                                self.LINE_END)
        form_data += to_utf8('Content-Disposition: form-data; name="%s"') % \
                                (name,)
        if file_name:
            form_data += to_utf8('; filename="%s"') % (file_name,)
        form_data += to_utf8(self.LINE_END)

        if type != None:
            form_data += to_utf8("Content-Type: %s%s") % \
                                (type, to_utf8(self.LINE_END))

        if data != None:
            form_data += to_utf8(self.LINE_END) + data + to_utf8(self.LINE_END)
        return form_data

    def _get_prefix_string(self, file_name):
        """Return the prefix string data for uploading."""

        prefix = self._get_form_data_string(to_utf8("attributes"),
                            to_utf8(UPLOAD_ATTRIBUTES_TMPL) % \
                                (file_name, self.parent_id))
        prefix += self._get_form_data_string(to_utf8(self.FILE_NAME),
                                type=to_utf8("application/octet-stream"),
                                file_name=file_name)

        prefix += to_utf8(self.LINE_END)
        return prefix

    def _get_suffix_string(self):
        """Return the suffix string data for uploading."""

        return to_utf8(self.LINE_END + self.BOUNDARY_PREFIX + self.boundary + \
                        self.BOUNDARY_SUFFIX + self.LINE_END)

    def _read(self, stream, size=-1):
        """Return the data in "stream"."""

        data = to_utf8("")
        read_size = size
        while True:
            buf = stream.read(read_size)
            if len(buf) <= 0:
                return data
            data += buf
            if size > 0:
                read_size = size - len(buf)

    def __init__(self, file_name, parent_id, data, length, boundary):
        """Constructor."""

        file_name = to_utf8(file_name)
        boundary = to_utf8(boundary)
        self.parent_id = to_utf8(parent_id)
        self.boundary = boundary
        self.headers = {to_utf8("Content-Type"):
                    to_utf8("multipart/form-data; boundary=%s") % (boundary,)}
        self.prefix = DataStream(self._get_prefix_string(file_name))
        self.main = DataStream(data, length)
        self.suffix = DataStream(self._get_suffix_string())

        self.stream_list = (self.prefix, self.main, self.suffix)
        self.length = self.prefix.length() + self.main.length() + \
                        self.suffix.length()
        self.headers[to_utf8("Content-length")] = self.length
        self.reset_stream()

    def read(self, size=-1):
        """Return the upload data."""

        data = to_utf8("")
        read_size = size
        for stream in self.stream_list:
            data += self._read(stream, read_size)
            if size >= 0 and len(data) == size:
                return data

            if size > 0:
                read_size = size - len(data)

        return data


class BoxFileObj(FileObj):

    def _write_from_path(self, url, local_path):

        with open(local_path, "r") as local_file_obj:
            upload_file_data = BoxUploadFileData(
                        to_utf8(os.path.basename(self._path)),
                        self._client._get_id(os.path.dirname(self._path)),
                        local_file_obj, self._length, BOUNDARY_MAIN)
            return self._client._urlopen(url, upload_file_data,
                                add_headers=upload_file_data.get_headers())

    def __init__(self, client, path, mode, length, _file=None):

        super(BoxFileObj, self).__init__()
        self._client = client
        self._path = path
        self._file = _file
        self._mode = mode
        self._length = length
        self._write_data = ""
        self._offset = 0

        if (not self._mode == OPEN_MODE_READ and
            not self._mode == OPEN_MODE_WRITE):

            raise ValueError(ERR_MSG_INVALID_OPEN_MODE)

        if self._mode == OPEN_MODE_READ:
            url = self._client._base_url + FILE_READ_URL_TMPL % \
                                        (self._client._get_id(self._path),)
            self._file = self._client._urlopen(url)

    def close(self):

        if self._file is not None:
            del self._file
            self._file = None

    def read(self, size=-1):

        return self._file.read(size)

    def write_from_path(self, local_path):

        if isinstance(local_path, unicode):
            local_path = local_path.encode("UTF-8")

        try:
            url = self._client._upload_base_url + FILE_WRITE_URL_TMPL
            response = self._write_from_path(url, local_path)
            self._client._update_path_id_map(self._path,
                                json.loads(response.read()))
        except urllib2.HTTPError, err:
            if err.code != httplib.CONFLICT:
                raise

            err_data = err.read()
            err_info = ErrorInfo(to_json(err_data))
            if not err_info.id or err_info.code != ErrorInfo.ITEM_NAME_IN_USE:
                raise recreate_http_error(err, err_data)

            url = self._client._upload_base_url + \
                    FILE_UPDATE_URL_TMPL % (err_info.id,)
            self._write_from_path(url, local_path)


class BoxAccess(Access):

    def _to_unicode(self, target):

        if isinstance(target, str):
            return target.decode("utf8")
        return target

    def _to_mtime(self, mtime_str):
        """Transform "mtime_str" to a UTC epoch time and return it."""

        if not mtime_str:
            return 0

        match = re.match(MTIME_UTC_RE, mtime_str)
        tz_adjust = 0
        if match == None:
            match = re.match(MTIME_TZ_RE, mtime_str)
            tz_adjust = int(match.group(3)) * 3600 + int(match.group(4)) * 60
            if match.group(2) == "+":
                tz_adjust *= -1

        time_tuple = time.strptime(match.group(1), MTIME_FORMAT)
        return int(calendar.timegm(time_tuple)) + tz_adjust

    def _to_stat_info(self, content_json):
        """Return stat dictionary which created based on "content_json"."""

        is_dir = content_json[TYPE_NAME] == FOLDER_TYPE
        if is_dir:
            mode = stat.S_IFDIR
            size = 0
        else:
            mode = 0
            size = content_json.get(SIZE_NAME, 0)

        mtime = self._to_mtime(content_json[MODIFIED_AT_NAME])

        return {
                stat.ST_MODE: mode,
                stat.ST_MTIME: mtime,
                stat.ST_SIZE: size,
            }

    def _get_headers(self):
        """Return dictionary of headers."""

        headers = {}
        if ACCESS_TOKEN_NAME in self._access_token_info:
            headers["Authorization"] = "Bearer " + \
                self._access_token_info[ACCESS_TOKEN_NAME]
        if self._http_user_agent:
            headers["User-Agent"] = self._http_user_agent

        return headers

    def _handle_unauthorized(self, retry_count):
        """Handle HTTP UNAUTHORIZED error."""

        if retry_count < UNAUTHORIZED_RETRY_COUNT:
            self._refresh_access_token_info()
            raise NeedToRetry()

    def _handle_exception(self, exc, retry_count):
        """Handle "exc" exception."""

        try:
            raise exc
        except urllib2.HTTPError, err:
            if (err.code == httplib.INTERNAL_SERVER_ERROR or
                err.code == httplib.BAD_GATEWAY or
                err.code == httplib.SERVICE_UNAVAILABLE or
                err.code == httplib.GATEWAY_TIMEOUT):

                if retry_count < SERVER_ERR_RETRY_COUNT:
                    time.sleep(RETRY_SLEEP_TIME)
                    raise NeedToRetry()
                raise

            err_data = err.read()

            if err.code == httplib.UNAUTHORIZED:
                self._handle_unauthorized(retry_count)
            raise recreate_http_error(err, err_data)

    def _urlopen(self, url, data=None, method=None, add_headers={}):
        """Open "url" and return the file like object."""

        count = 0

        while True:
            try:
                headers = self._get_headers()
                headers.update(add_headers)
                request = urllib2.Request(url, data, headers)
                if method:
                    request.get_method = lambda: method
                if self.opener != None:
                    return self.opener.open(request)

                return urllib2.urlopen(request)
            except:
                try:
                    self._handle_exception(sys.exc_info()[1], count)
                    # Do not come here
                except NeedToRetry:
                    count += 1
                    continue

    def _save_access_token_info(self):
        """Save current access token info to session file."""

        with open(self._session_path, "w") as session_file:
            session_file.write(json.dumps(self._access_token_info))
        os.chmod(self._session_path, SESSION_PATH_PERM)

    def _load_access_token_info(self):
        """Load access token info from session file."""

        with open(self._session_path) as session_file:
            self._access_token_info = json.loads(session_file.read())

    def _get_access_token_info(self, code):
        """Return a access token data for the content API of BOX server."""

        param = urlencode({
                "grant_type": "authorization_code",
                "code": code,
                "client_id": self._client_id,
                "client_secret": self._client_secret,
                "redirect_url": self._redirect_url,
            })

        response = self._urlopen(self._oauth_base_url + "/token", param)
        return json.loads(response.read())

    def _refresh_access_token_info(self):
        """Refresh a access token and return it."""

        param = urlencode({
                "grant_type": "refresh_token",
                "refresh_token": self._access_token_info[REFRESH_TOKEN_NAME],
                "client_id": self._client_id,
                "client_secret": self._client_secret,
            })

        response = self._urlopen(self._oauth_base_url + "/token", param)
        self._access_token_info = json.loads(response.read())
        self._save_access_token_info()

    def _get_id_dict(self, contents):
        """Return dictionary of the id in "contents"."""

        id_dict = {}
        for content in contents.get(ENTRIES_NAME, []):
            id = content.get(ID_NAME, "")
            name = content.get(NAME_NAME, "")
            if id and name:
                id_dict[name] = id

        return id_dict

    def _normalize_path(self, path):
        """Delete a first "/" if path starts with "//"."""

        if path == self.sep:
            return path
        path = path.rstrip(self.sep)
        if path.startswith("//"):
            return path[1:]
        return path

    def _update_path_id_map(self, path, contents):
        """Update self._path_id_map with information in "response"."""

        path = self._normalize_path(path)
        id = ""
        if ENTRIES_NAME in contents:
            name = os.path.basename(path)
            id_dict = self._get_id_dict(contents)
            if name in id_dict:
                id = id_dict[name]
        else:
            id = contents.get(ID_NAME, "")

        if id:
            self._path_id_map.append_path(path, id)

    def _set_handler(self, proxy=None):

        handlers = []
        if hasattr(ssl, SSL_CREATE_UNVERIFIED_CONTEXT):
            handlers += [urllib2.HTTPSHandler(
                                    context=ssl._create_unverified_context())]
        if proxy != None:
            proxies = get_proxies_dict(proxy, PROTOCOL_LIST)
            if proxies:
                handlers += [urllib2.ProxyHandler(proxies)]

        if handlers:
            self.opener = urllib2.build_opener(*handlers)
        else:
            self.opener = None

    def _get_id(self, target_path):

        try:
            target_id = None
            target_id = self._path_id_map.get_id(target_path)
            if target_id is not None:
                return target_id
        except LookupError:
            pass

        parent_path = self._normalize_path(os.path.dirname(target_path))
        parent_id = self._get_id(parent_path)
        if parent_id is None:
            raise LookupError(target_path)

        url = self._base_url + FOLDERS_ITEMS_URL_TMPL % (parent_id,)
        offset = 0
        limit = ITEM_LIMIT
        while True:
            url += "?" + FOLDERS_ITEMS_PARAM_TMPL % (limit, offset)
            response = self._urlopen(url)
            response_json = json.loads(response.read())
            for content in response_json.get(ENTRIES_NAME, []):
                name = self._to_unicode(content[NAME_NAME])
                id = content.get(ID_NAME, "")
                if not id:
                    continue
                content_path = os.path.join(parent_path, name)
                if target_path == content_path:
                    target_id = id
                self._path_id_map.append_path(content_path, id)

            offset += limit
            if offset > response_json[TOTAL_COUNT_NAME]:
                break

        if not target_id:
            raise LookupError(target_path)

        return target_id

    def __init__(self, client_id, client_secret, base_url, upload_base_url,
            oauth_base_url, redirect_url, session_path=None, proxy=None,
            http_user_agent=None):

        self.sep = "/"

        self._client_id = client_id
        self._client_secret = client_secret
        self._base_url = base_url
        self._upload_base_url = upload_base_url
        self._oauth_base_url = oauth_base_url
        self._redirect_url = redirect_url
        self._session_path = session_path
        self._http_user_agent = http_user_agent
        self._proxy = proxy
        self.opener = None
        self._access_token_info = {}

        self._path_id_map = PathIdMap()
        self._path_id_map.append_path(ROOT_PATH, ROOT_FOLDER_ID)

        self._set_handler(self._proxy)

    @property
    def auth_url(self):
        """Get the auth url."""

        param = {
            "response_type": "code",
            "client_id": self._client_id,
            "redirect_url": self._redirect_url,
        }

        return self._oauth_base_url + "/authorize" + "?" + urlencode(param)

    def authenticate(self, params=None):
        """Takes in params, creates a session.

        Args:
            params(dict):
                The GET parameter sent to redirect_url
                If session is already valid, is None.
                e.g.
                {'code': '....'}
        """

        if params is None:
            params = {}

        code = params.get('code')
        if code is not None:
            self._access_token_info = self._get_access_token_info(code)
            self._save_access_token_info()
        else:
            self._load_access_token_info()

    def make_directory(self, dir_path):
        """Create "dir_path" directory."""

        dir_path = self._normalize_path(dir_path)
        param = json.dumps({
                NAME_NAME: to_utf8(os.path.basename(dir_path)),
                PARENT_NAME: {
                        ID_NAME: self._get_id(os.path.dirname(dir_path)),
                    }
            })
        try:
            response = self._urlopen(self._base_url + MAKE_DIRECTORY_URL_TMPL,
                                        param)
            self._update_path_id_map(dir_path, json.loads(response.read()))
        except urllib2.HTTPError, err:
            if err.code != httplib.CONFLICT:
                raise
            # Assume it already exists and there is no error.

    def delete_directory(self, dir_path):
        """Delete "dir_path" directory."""

        dir_path = self._normalize_path(dir_path)
        url = self._base_url + FOLDERS_ID_URL_TMPL % (self._get_id(dir_path),)
        url += "?" + RECURSIVE_PARAM
        self._urlopen(url, method=DELETE_METHOD)
        self._path_id_map.delete_path(dir_path)

    def stat_directory(self, dir_path):
        """stat "dir_path" directory."""

        dir_path = self._normalize_path(dir_path)
        url = self._base_url + FOLDERS_ID_URL_TMPL % (self._get_id(dir_path),)
        response = self._urlopen(url)
        return self._to_stat_info(json.loads(response.read()))

    def rename_directory(self, from_path, new_name):

        from_path = self._normalize_path(from_path)
        url = self._base_url + FOLDERS_ID_URL_TMPL % (self._get_id(from_path),)
        param = json.dumps({
                NAME_NAME: to_utf8(new_name),
            })
        self._urlopen(url, param, method=PUT_METHOD)

    def move_directory(self, from_path, to_path):

        from_path = self._normalize_path(from_path)
        to_path = self._normalize_path(to_path)
        url = self._base_url + FOLDERS_ID_URL_TMPL % (self._get_id(from_path),)
        param = json.dumps({
                PARENT_NAME: {
                        ID_NAME: self._get_id(to_path),
                    }
            })
        self._urlopen(url, param, method=PUT_METHOD)

    def get_contents(self, dir_path, with_stat=False):

        dir_path = self._normalize_path(dir_path)
        dir_id = self._get_id(dir_path)
        self._path_id_map.append_directory(dir_path, dir_id)

        url = self._base_url + FOLDERS_ITEMS_URL_TMPL % (dir_id,)
        offset = 0
        limit = ITEM_LIMIT
        contents = {}
        while True:
            url += "?" + FOLDERS_ITEMS_PARAM_TMPL % (limit, offset)
            response = self._urlopen(url)
            response_json = json.loads(response.read())
            for content in response_json.get(ENTRIES_NAME, []):
                name = self._to_unicode(content[NAME_NAME])
                if with_stat:
                    contents[name] = self._to_stat_info(content)
                else:
                    contents[name] = {}

                id = content.get(ID_NAME, "")
                if not id:
                    continue
                content_path = os.path.join(dir_path, name)
                self._path_id_map.append_path(content_path, id)

            offset += limit
            if offset > response_json[TOTAL_COUNT_NAME]:
                break

        return contents

    def open(self, file_path, mode="r", length=-1, chunk_size=None):
        """Return a class instance of the "FileObj" or sub class
           corresponding to "file_path".
        """

        file_path = self._normalize_path(file_path)
        return BoxFileObj(self, file_path, mode, length)

    def delete_file(self, file_path):

        file_path = self._normalize_path(file_path)
        url = self._base_url + FILES_ID_URL_TMPL % (self._get_id(file_path),)
        self._urlopen(url, method=DELETE_METHOD)
        self._path_id_map.delete_path(file_path)

    def stat_file(self, file_path):

        file_path = self._normalize_path(file_path)
        url = self._base_url + FILES_ID_URL_TMPL % (self._get_id(file_path),)
        response = self._urlopen(url)
        return self._to_stat_info(json.loads(response.read()))

    def rename_file(self, from_path, new_name):

        from_path = self._normalize_path(from_path)
        url = self._base_url + FILES_ID_URL_TMPL % (self._get_id(from_path),)
        param = json.dumps({
                NAME_NAME: to_utf8(new_name),
            })
        self._urlopen(url, param, method=PUT_METHOD)

    def move_file(self, from_path, to_path):

        from_path = self._normalize_path(from_path)
        to_path = self._normalize_path(to_path)
        url = self._base_url + FILES_ID_URL_TMPL % (self._get_id(from_path),)
        param = json.dumps({
                PARENT_NAME: {
                        ID_NAME: self._get_id(to_path),
                    }
            })
        self._urlopen(url, param, method=PUT_METHOD)
